@@ -4,8 +4,8 @@ from __future__ import print_function, unicode_literals
 
 about = {
     "name": "defrostir",
-    "version": "0.15",
-    "date": "2021-10-28",
+    "version": "1.0",
+    "date": "2023-10-31",
     "description": "split broken icecast recordings into separate mp3s",
     "author": "ed",
     "license": "MIT",
@@ -25,12 +25,12 @@ import struct
 import hashlib
 import logging
 import argparse
+import calendar
 import platform
 import threading
 import traceback
 import unicodedata
 import subprocess as sp
-from datetime import datetime
 
 try:
     from itertools import zip_longest
@@ -98,6 +98,35 @@ YOLO = False  # cfg: True disengages safety checks
 METAINT = 16000  # cfg: should match the headers from the icecast server
 IO_BUFSZ = 512 * 1024  # cfg: read/write buffer size, probably optimal
 
+
+try:
+    from datetime import datetime, timezone
+
+    UTC = timezone.utc
+except ImportError:
+    from datetime import datetime, timedelta, tzinfo
+
+    TD_ZERO = timedelta(0)
+
+    class _UTC(tzinfo):
+        def utcoffset(self, dt):
+            return TD_ZERO
+
+        def tzname(self, dt):
+            return "UTC"
+
+        def dst(self, dt):
+            return TD_ZERO
+
+    UTC = _UTC()
+
+
+if not PY3:
+    import io
+
+    open = io.open
+
+
 if PY3 and not PYPY:
     uprint = print
 else:
@@ -128,14 +157,14 @@ class LoggerFmt(logging.Formatter):
         else:
             c = "1;31"
 
-        ts = datetime.utcfromtimestamp(record.created)
+        ts = datetime.fromtimestamp(record.created, UTC)
         ts = ts.strftime("%H:%M:%S.%f")[:-3]
 
         msg = record.msg
         if record.args:
             msg = msg % record.args
 
-        return "\033[0;36m{}\033[0{}m {}\033[0m".format(ts, c, msg)
+        return "\033[0;36m%s\033[0%sm %s\033[0m" % (ts, c, msg)
 
 
 def configure_logger(debug):
@@ -154,7 +183,7 @@ def configure_logger(debug):
 
 def reprint(*args, **kwargs):
     msg = args[0]
-    msg = "\033[A{}\033[K".format(msg)
+    msg = "\033[A%s\033[K" % (msg,)
     uprint(msg, *list(args)[1:], **kwargs)
 
 
@@ -309,7 +338,7 @@ def fmt_meta(metasrc_yield, sz, ntags):
     pos, buf = metasrc_yield
     meta, enc, err = tag2text(buf)
 
-    return "at {}/{} ({:.2f}%), {} tags, [{}] [{}] {}".format(
+    return "at %s/%s (%.2f%%), %s tags, [%s] [%s] %s" % (
         pos, sz, (pos * 100.0 / sz), ntags, enc, meta, err
     )
 
@@ -326,10 +355,7 @@ def ensure_parsers_agree(metasrc1, metasrc2, sz):
 
         if err:
             msg = "scanner/verifier disagreement,\n{} [{}]\n{} [{}]".format(
-                pos1,
-                tag2text(buf1)[0],
-                pos2,
-                tag2text(buf2)[0],
+                pos1, tag2text(buf1)[0], pos2, tag2text(buf2)[0],
             )
             msg += "\nthis is not an error (should just be ignored) "
             msg += "but raising for inspection until properly tested"
@@ -456,7 +482,7 @@ def defrost(fn, metasrc, sz):
 
 def run_defrost(fn, fn_mp3, fn_idx, sz):
     with open(fn, "rb", IO_BUFSZ) as fi, open(fn_mp3, "wb", IO_BUFSZ) as fom, open(
-        fn_idx, "w", IO_BUFSZ
+        fn_idx, "w", IO_BUFSZ, encoding="utf-8"
     ) as foi, IcyScanner(fn) as metasrc:
         ntags = 0
         fom_pos = 0
@@ -528,14 +554,6 @@ def run_defrost(fn, fn_mp3, fn_idx, sz):
 
 
 def collect_frames(fn_mp3):
-    # cmd /c "set FFREPORT=level=32:file=ffprobe.log & ffprobe -hide_banner -select_streams a:0 -show_frames -show_entries frame=pkt_pts_time,pkt_dts_time,best_effort_timestamp_time,pkt_duration_time,pkt_pos,nb_samples -of compact=p=0 5g-ok.mp3.defrost.mp3 >ffprobe.out"
-    # tail -F ffprobe.out | awk 'NR%1000==1'
-    # perl -e 'while(<>){if (!/^pkt_pts_time=([0-9\.]+)\|pkt_dts_time=([0-9\.]+)\|best_effort_timestamp_time=([0-9\.]+)\|pkt_duration_time=([0-9\.]+)\|pkt_pos=([0-9]+)\|nb_samples=([0-9]+)\r?$/) {print$_;next} printf "%.3f %.3f %.3f %.3f %d %d\n",$1,$2,$3,$4,$5,$6}' <ffprobe.out >ffprobe.outf
-
-    ptn = r"^pkt_pts_time=([0-9\.]+)\|pkt_dts_time=([0-9\.]+)\|best_effort_timestamp_time=([0-9\.]+)\|pkt_pos=([0-9]+)\r?$"
-    # ptn = r"^pts_time=([0-9\.]+)\|dts_time=([0-9\.]+)\|pos=([0-9]+)\r?$"
-    ptn = re.compile(ptn)
-
     # fmt: off
     c = [
         "ffprobe",
@@ -543,7 +561,7 @@ def collect_frames(fn_mp3):
         "-select_streams", "a:0",
         "-show_frames",
         # "-show_packets",
-        "-show_entries", "frame=pkt_pts_time,pkt_dts_time,best_effort_timestamp_time,pkt_pos",
+        "-show_entries", "frame=pkt_pts_time,pkt_dts_time,best_effort_timestamp_time,pkt_pos,channels",
         # "-show_entries", "packet=pts_time,dts_time,pos",
         "-of", "compact=p=0",
         fn_mp3,
@@ -559,8 +577,10 @@ def collect_frames(fn_mp3):
     while True:
         ibuf = p.stdout.read(4096)
         if not ibuf:
-            if p.poll() is not None:
-                info("FFmpeg terminated")
+            rc = p.poll()
+            if rc is not None:
+                fun = warn if rc else info
+                fun("FFmpeg exited with status %d", rc)
                 return
             else:
                 fails += 1
@@ -579,19 +599,26 @@ def collect_frames(fn_mp3):
 
         sbuf = tbuf.decode("utf-8")
         for ln in sbuf.split("\n"):
-            m = ptn.match(ln)
-            if not m:
+            try:
+                fields = [x.split("=") for x in ln.split("|")]
+                fields = {k: v for k, v in fields}
+                _ = fields["best_effort_timestamp_time"]
+            except:
                 continue
 
-            pts, dts, ts, pos = m.groups()
-            if pts != dts or pts != ts:
+            pts = fields.get("pkt_pts_time")
+            dts = fields.get("pkt_dts_time")
+            ts = fields.get("best_effort_timestamp_time")
+            if (
+                (pts and dts and pts != dts)
+                or (pts and ts and pts != ts)
+                or (dts and ts and dts != ts)
+            ):
                 warn("defrost: spooky timestamps ({},{},{})\n".format(pts, dts, ts))
 
-            # ts, dts, pos = m.groups()
-            # if ts != dts:
-            #     warn("defrost: spooky timestamps ({},{})\n".format(ts, dts))
+            ts = ts or dts or pts
 
-            yield [int(pos), float(ts)]
+            yield [int(fields["pkt_pos"]), float(ts)]
 
 
 def detect_silence_one(fn_mp3, db, len_sec, xpos, ret):
@@ -708,10 +735,7 @@ def detect_silence_one(fn_mp3, db, len_sec, xpos, ret):
 def build_framecache(f_frames, ofs1, ofs2):
     lower_byte = max(0, ofs1 - 320 * 128 * 60)
     extra_sec = 3600
-    msg = "rebuild framecache {}..+{} for {}..{}".format(
-        lower_byte, extra_sec, ofs1, ofs2
-    )
-    debug(msg)
+    debug("rebuild framecache %s..+%s for %s..%s", lower_byte, extra_sec, ofs1, ofs2)
     nframe = 0
     framecache = []
 
@@ -883,7 +907,7 @@ def tag_mp3(fn, tagtxt, timestr, album):
         id3.add(COMM(encoding=3, text=comment))
         id3.save(fn)
     except Exception as ex:
-        warn("failed to write tags: {}".format(repr(ex)))
+        warn("failed to write tags: %r", ex)
 
 
 def add_sentinel(gen):
@@ -903,7 +927,7 @@ def main():
     ap.add_argument("-f", action="store_true", help="overwrite existing split")
     ap.add_argument("-o", metavar="DIR", help="output directory")
     ap.add_argument("-a", metavar="ALBUM", help="album title for id3 tags")
-    ap.add_argument("--metaint", type=int, help="icecast metadata interval", default=METAINT)
+    ap.add_argument("--metaint", metavar="N", type=int, help="icecast metadata interval in bytes; default %s" % (METAINT,), default=METAINT)
     ap.add_argument("--no-split", action="store_true", help="do not split the mp3")
     ap.add_argument("--no-id3", action="store_true", help="do not write id3 tags")
     ap.add_argument("--yolo", action="store_true", help="less sanity checks (for buggy files)")
@@ -1004,11 +1028,11 @@ def main():
         run_defrost(fn, fn_mp3, fn_idx, sz)
 
     if os.path.exists(fn_frames):
-        with open(fn_frames, "r") as f:
+        with open(fn_frames, "r", encoding="utf-8") as f:
             ln = next(f)
             fields = ln.split(" ")
             if len(fields) != 3 or fields[0] != "0":
-                warn("deleting old {} (format has changed)".format(fn_frames))
+                warn("deleting old %s (format has changed)", fn_frames)
                 os.unlink(fn_frames)
 
     sz = os.path.getsize(fn_mp3)
@@ -1018,12 +1042,12 @@ def main():
         buf = []
         pos = 0
         nframes = -1
-        with open(fn_frames, "w") as f:
+        with open(fn_frames, "w", encoding="utf-8") as f:
             for pos, ts in collect_frames(fn_mp3):
                 nframes += 1
-                buf.append("{} {} {:.3f}\n".format(nframes, pos, ts))
+                buf.append("%s %s %.3f\n" % (nframes, pos, ts))
                 if len(buf) > 1024 * 16:
-                    print("\033[A{:.2f}% ".format(pos * 100.0 / sz))
+                    print("\033[A%.2f%% " % (pos * 100.0 / sz,))
                     f.write("".join(buf))
                     buf = []
             if buf:
@@ -1036,7 +1060,7 @@ def main():
 
     print("\n")
     info("reading eof from frametab")
-    with open(fn_frames, "r", IO_BUFSZ) as f:
+    with open(fn_frames, "r", IO_BUFSZ, encoding="utf-8") as f:
         f.seek(0, os.SEEK_END)
         f.seek(f.tell() - 1024)
         next(f)
@@ -1069,16 +1093,16 @@ def main():
             r.extend(task[1])
 
         r.sort()
-        with open(fn_silence, "w") as f:
+        with open(fn_silence, "w", encoding="utf-8") as f:
             for nframe, db, start, end in r:
-                ln = "{} {} {:.2f} {:.2f}\n".format(nframe, db, start, end)
+                ln = "%s %s %.2f %.2f\n" % (nframe, db, start, end)
                 f.write(ln)
 
     # probably small enough
     print()
     info("reading silent ranges into memory")
     silents = {}  # dB => [nframe, t1, t2]  (prefer timestamps)
-    with open(fn_silence, "r") as f:
+    with open(fn_silence, "r", encoding="utf-8") as f:
         for ln in f:
             fields = ln.split(" ")
             nframe, db = [int(x) for x in fields[:2]]
@@ -1092,9 +1116,13 @@ def main():
     if ar.i == "ls":
         # {"o": 6592000, "o1": 6608493, "o2": 12816880, "t": "Q2xhcmlTIC0gaXJvbnk=", "z": [16000, 32001, 48002,
         ptn = re.compile(r"^([0-9]{2,8}):([0-9]{2}):([0-9]{2}) (.*)$")
-        with open(
-            fn + ".txt", "r", encoding="utf-8-sig", errors="replace"
-        ) as f_txt, open(fn_frames, "r") as f_frames, open(fn_idx, "w") as f_idx:
+        with open(fn + ".txt", "r", encoding="utf-8", errors="replace") as f_txt, open(fn_frames, "r", encoding="utf-8") as f_frames, open(fn_idx, "w", encoding="utf-8") as f_idx:
+            for _ in range(4096):
+                ofs = f_txt.tell()
+                if f_txt.read(1) in "1234567890":
+                    break
+            f_txt.seek(ofs)
+
             for txt_ln in f_txt:
                 m = ptn.match(txt_ln)
                 if not m:
@@ -1117,7 +1145,7 @@ def main():
 
     ntrack = 0
     framecache = []  # nframe, ofs, sec
-    with open(fn_idx, "r") as f_idx, open(fn_frames, "r") as f_frames, open(
+    with open(fn_idx, "r", encoding="utf-8") as f_idx, open(fn_frames, "r", encoding="utf-8") as f_frames, open(
         fn_mp3, "rb"
     ) as f_defrosted:
         tag = None
@@ -1140,7 +1168,7 @@ def main():
             ofs1 = tag["o"]
             ofs2 = tag2["o"]
             perc = ofs1 * 100.0 / sz
-            debug("{:.0f}% track #{}, ofs {}..{}".format(perc, ntrack, ofs1, ofs2))
+            debug("%.0f%% track #%s, ofs %s..%s", perc, ntrack, ofs1, ofs2)
 
             # 38 frames in 1 sec (1152 samples per frame),
             # 192k=627b, 256k=836bm 320k=1045b
@@ -1184,10 +1212,8 @@ def main():
                 msg = "could not find next_sec {} in framecache, min is {} (#{}, {}b), gave up at {} (#{}, {}b)"
                 raise Exception(msg.format(next_sec, c, a, b, ts, nframe, ofs))
 
-            msg = "{:.0f}% track #{}, ofs {}..{}, sec {}..{}, frame {}..{}"
-            debug(
-                msg.format(perc, ntrack, ofs1, ofs2, tag_ts1, tag_ts2, frame1, frame2)
-            )
+            msg = "%.0f%% track #%s, ofs %s..%s, sec %s..%s, frame %s..%s"
+            debug(msg, perc, ntrack, ofs1, ofs2, tag_ts1, tag_ts2, frame1, frame2)
 
             ts1 = tag_ts1
             ts2 = tag_ts2
@@ -1200,6 +1226,12 @@ def main():
                 quiet_ts_ref = tag_ts2
                 quiet_frame2 = frame2
 
+            if quiet_ts2 is None and ofs1 > sz - 4096:
+                t = "tag at time %s, byte %s, is past mp3 eof %s: %s"
+                warn(t, tag_ts1, ofs1, sz, tag2text(tag["t"]))
+                rc += 1
+                continue  # or maybe break
+
             for nframe, ofs, ts in framecache:
                 if ts > quiet_ts2:
                     break
@@ -1249,12 +1281,18 @@ def main():
 
             tagbin = tag["t"]
             tagtxt, enc, _ = tag2text(tagbin)
-            debug("{:.0f}% tag: [{}] [{}]".format(perc, enc, tagtxt))
+            debug("%.0f%% tag: [%s] [%s]", perc, enc, tagtxt)
 
             if not ar.no_split:
                 unix = int(lastmod - (end_ts - ts1))
+                if ar.i == "ls":
+                    ptn = r"([0-9]{4}-[0-9]{2}-[0-9]{2}_[0-9]{2}\.[0-9]{2}\.[0-9]{2})"
+                    m = re.search(ptn, os.path.basename(ar.ICY_MP3))
+                    if m:
+                        dt = time.strptime(m.group(1), "%Y-%m-%d_%H.%M.%S")
+                        unix = int(calendar.timegm(dt)) + ts1
                 fmt = "%Y-%m-%d, %H:%M:%S"
-                timestr = datetime.utcfromtimestamp(unix).strftime(fmt)
+                timestr = datetime.fromtimestamp(unix, UTC).strftime(fmt)
                 fn = split_mp3(f_defrosted, ntrack, ofs1, ofs2, outdir, tagtxt)
                 if not ar.no_id3:
                     tag_mp3(fn, tagtxt, timestr, ar.a)
@@ -1302,6 +1340,12 @@ if __name__ == "__main__":
 ## start of usage notes
 
 
+##
+## convert a folder of loopstream recordings starting at file "Loopstream-2021-10-31_21.57.26.mp3"
+
+# en=; for f in *.mp3; do [ "$f" = "Loopstream-2021-10-31_21.57.26.mp3" ] && en=1; [ $en ] || continue; t=${f:11:4}-${f:16:2}${f:19:2}-${f:22:2}${f:25:2}${f:28:2}; echo $t; PYTHONPATH=/c/Users/ed/dev/defrost/ python -m defrost -i ls -o $t "$f"; done; rm -- *.mp3.defrost-*
+
+
 ##
 ## get all dj changes and songs played from znc irc logs
 
